from __future__ import annotations
from dataclasses import dataclass
from enum import auto, Enum, IntEnum
from functools import cmp_to_key, total_ordering
from itertools import combinations, groupby
from operator import attrgetter, itemgetter
from collections import defaultdict
from collections.abc import Iterable

class Suit(Enum):
   Hearts = auto()
   Spades = auto()
   Diamonds = auto()
   Clubs = auto()

@dataclass(frozen=True)
class Card:
   suit: Suit
   rank: int

class HandType(IntEnum):
   HighCard = auto()
   Pair = auto()
   TwoPair = auto()
   ThreeOfAKind = auto()
   Straight = auto()
   Flush = auto()
   FullHouse = auto()
   FourOfAKind = auto()
   StraightFlush = auto()

class Hand:
   _type: HandType
   _cards: list[tuple[int, int]]

   @staticmethod
   def getHand(cards: Iterable[Card]) -> Hand:
      groups = sorted(
         ((g[0], len(list(g[1]))) for g in groupby(sorted(cards, key=attrgetter('rank')), key=attrgetter('rank'))),
         key=cmp_to_key(lambda a, b: a[1] - b[1] if a[1] != b[1] else a[0] - b[0]), reverse=True)

      match [g[1] for g in groups]:
         case [4, 1]:
            return Hand(HandType.FourOfAKind, groups)
         case [3, 2]:
            return Hand(HandType.FullHouse, groups)
         case [3, 1, 1]:
            return Hand(HandType.ThreeOfAKind, groups)
         case [2, 2, 1]:
            return Hand(HandType.TwoPair, groups)
         case [2, 1, 1, 1]:
            return Hand(HandType.Pair, groups)
         case _:
            flush = len(set(card.suit for card in cards)) == 1
            straight = (groups[1][0] == groups[0][0] - 1 and
               groups[2][0] == groups[0][0] - 2 and
               groups[3][0] == groups[0][0] - 3 and
               groups[4][0] == groups[0][0] - 4)

            if straight and flush:
               return Hand(HandType.StraightFlush, groups)
            elif straight:
               return Hand(HandType.Straight, groups)
            elif flush:
               return Hand(HandType.Flush, groups)
            else:
               return Hand(HandType.HighCard, groups)

   @staticmethod
   def getBestHand(cards: Iterable[Card]):
      ace14 = [Card(card.suit, 14) if card.rank == 1 else card for card in cards]

      return max(Hand.getHand(cards), Hand.getHand(ace14))
   
   def __init__(self: Hand, type: HandType, cards: list[list[Card]]):
      self._type = type
      self._cards = cards

   def __lt__(self: Hand, other: Hand):
      if (self._type != other._type):
         return self._type < other._type
      else:
         for i, j in zip(self._cards, other._cards):
            if i != j:
               return i < j


def bestHand(player_cards: list[Card], community_cards: list[Card]):
   return max(Hand.getBestHand(player_cards + list(community)) for community in combinations(community_cards, 3))

def entry(players: dict[str, list[Card]], community_cards: list[Card]):
   remaining_cards = []
   for suit in Suit:
      for rank in range(1, 14):
         card = Card(suit, rank)
         if card not in community_cards and not any(card in hand for hand in players.values()):
            remaining_cards.append(card)

   remaining_cards_count = 5 - len(community_cards)
   
   n = 0
   win_count = defaultdict(lambda: 0)
   for community_cards2 in combinations(remaining_cards, remaining_cards_count):
      n += 1
      hands = sorted(((player, bestHand(players[player], community_cards + list(community_cards2))) for player in players),
         key=itemgetter(1), reverse=True)
      if hands[0][1] > hands[1][1]:
         win_count[hands[0][0]] += 1

   return { p: win_count[p] / n for p in win_count }

def cardsFromStr(str: str) -> list[Card]:
   suit_map = { 'H': Suit.Hearts, 'S': Suit.Spades, 'D': Suit.Diamonds, 'C': Suit.Clubs }
   value_map = { 'K': 13, 'Q': 12, 'J': 11, 'T': 10, 'A': 1 }
   return list(Card(suit_map[s[1]], value_map[s[0]] if s[0] in value_map else int(s[0])) for s in str.split(' '))

if __name__ == '__main__':
   players = {
      'paul': cardsFromStr('9H 3D'),
      'vanessa': cardsFromStr('6D 3C')
   }

   print(entry(players, cardsFromStr('2D 9S 9C')))
   print(entry(players, cardsFromStr('2D 9S 9C 4H')))
   print(entry(players, cardsFromStr('2D 9S 9C 4H 5C')))